function G = computeCmatrix(G, sigmastar, varargin)
% Compute relevante quantities for the curved mimetic method.
% One must first run computeGeomtry.m
% 
% SYNOPSIS:
%   G = computeCmatrix(G)
%   G = computeCmatrix(G, sigmastar)
%   G = computeCmatrix(G, sigmastar, forceOnFlat, useCurvedBC)
%
% PARAMETERS:
%   G               - Grid structure as described by grid_structure.
%
%   sigmastar       - threshold level for which faces are treated as
%                     strongly curved. Default=0.
%
%   forceOnFlat     - logical
%                       true:   use extra degrees of freedom for all faces,
%                               also for flat faces.
%                       false:  (default) extra degrees of freedom on faces
%                               strongly curved faces only.
%
%   useCurvedBC     - logical
%                       true:   use the curved mimetic boundary condition
%                       false:  (default) treat all boundary faces as
%                               moderately curved
%
% RETURNS:
%   G - Grid structure with added fields:
%         - CMFD
%             - N           -- A 3*Nf-by-3 matrix where 
%                              Nf = size(G.cells.faces,1). Rhe first Nf  
%                              rows are the area weighted normal vectors 
%                              and next Nf rows are the first set of
%                              orthonormal vector and the last Nf rows are
%                              the last set of orthonormal vectors
%
%             - C           -- A 3*Nf-by-3 matrix where the Nf first rows
%                              are the face centroids, and the next 2*Nf
%                              rows are constants relating to the
%                              orthonormal vectors and the centroids
%
%             - curvature   -- A Nf-by-1 array of the curvature constants
%                              sigma for each cell.
%
%         - faces
%             - cufg        -- A G.faces.num-by-1 array of logical values
%                              denoting moderately curved (false) and
%                              strongly curved (true) faces.
%                             
%
%             - curvature   -- A G.faces.num-by-1 array of the curvature 
%                              constants sigma for each cell.

%{
Copyright 2012 Ruben Bø.

This file is part of the Curved Mimetic Finite-Difference module (CMFD).
CMFD is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

CMFD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CMFD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CMFD.  If not, see <http://www.gnu.org/licenses/>.

%}

if nargin == 1
    sigmastar = 0;
    forceOnFlat = false;
    useCurvedBC = false;
elseif nargin == 2
    forceOnFlat = false;
    useCurvedBC = false;
elseif nargin == 4
    forceOnFlat = varargin{1};
    useCurvedBC = varargin{2};
else
    error('incorrect number of arguments')
end

% It is not allowed to use curvedBC when forceOnFlat is true
if forceOnFlat, assert(~useCurvedBC, 'forveOnFlat=true && useCurvedBC=true is not allowed'), end

% ------------------------------------------------------------------------
% Copied from the MRST file computeMimeticIP.m: Translate to local half-faces
    cf = double(G.cells.faces(:,1));
    diffN  = diff(G.faces.nodePos);
    nF     = length(cf);
    
    subcf = mcolon(G.faces.nodePos(cf), G.faces.nodePos(cf+1)-1);

    nsF    = length(subcf);

    cellno = rldecode(1 : G.cells.num, diff(G.cells.facePos), 2) .';
    sgn    = 2*double(G.faces.neighbors(cf, 1) == cellno) - 1;
    subsgn = rldecode(sgn,diffN(cf));
    
% ------------------------------------------------------------------------    
% Copied from the MRST file computeGeometry.m

    numf    = G.faces.num;
    faceNo  = rldecode(1:G.faces.num, diffN, 2) .';
    p       = G.faces.nodePos;
    next    = (2:size(G.faces.nodes, 1)+1) .';
    next(p(2 : end) - 1) = p(1 : end-1);
    llE = length(G.faces.nodes);
    localEdge2Face = sparse(1 : llE, faceNo, 1, llE, numf);
    sumUpOverSubFaces = sparse(1 : nsF, rldecode((1:nF)',diffN(cf)), 1, nsF, nF); 

    pCenters     = bsxfun(@rdivide, ...
                         localEdge2Face.' * G.nodes.coords(G.faces.nodes,:), ...
                         diff(double(G.faces.nodePos)));
    pCenters     = localEdge2Face * pCenters;
    subNormals   = cross(G.nodes.coords(G.faces.nodes(next),:) - ...
                        G.nodes.coords(G.faces.nodes,:), ...
                        pCenters - G.nodes.coords(G.faces.nodes,:)) ./ 2;
    subAreas     = sqrt(sum(subNormals .^ 2, 2));
    subCentroids = (G.nodes.coords(G.faces.nodes,:) + ...
                   G.nodes.coords(G.faces.nodes(next),:) + pCenters) ./ 3;
    faceAreas    = localEdge2Face.' * subAreas;

% ------------------------------------------------------------------------
% Compute the N matric, i.e., the orthogonals to the normal vectors 
% Using Gram-Schmidt to find the orthonormal vectors. Starting with e1 
% and e2 as a basis, choosing e3 if either is too close (by a factor 'tol')
    
    tol = 100*eps;
    
    N = G.faces.num;

    unitNormals = bsxfun(@rdivide,G.faces.normals,sqrt(sum(G.faces.normals.^2,2)));
        
    e1 = [ones(N,1)  zeros(N,1) zeros(N,1)];
    e2 = [zeros(N,1) ones(N,1)  zeros(N,1)];
    e3 = [zeros(N,1) zeros(N,1) ones(N,1) ];
    
    v2 = e1 - bsxfun(@times,unitNormals(:,1),unitNormals);
    normv1 = sqrt(sum(v2.^2,2));
    useX = normv1 >= tol;
    v2(useX,:) = bsxfun(@rdivide,bsxfun(@times,v2(useX,:),sign(v2(useX,1))),normv1(useX));
    
    if ~all(useX)
        w1 = e3(~useX,:) - bsxfun(@times,unitNormals(~useX,3),unitNormals(~useX,:));
        normw1 = sqrt(sum(w1.^2,2));
        assert(all(normw1 >= tol),'rounding error? maybe tol is too strikt?')
        w1 = bsxfun(@rdivide,bsxfun(@times,w1,sign(w1)),normw1);
        v2(~useX,:) = w1;
    end
      
    v3 = e2 - bsxfun(@times,unitNormals(:,2),unitNormals) - bsxfun(@times,v2(:,2),v2);
    normv2 = sqrt(sum(v3.^2,2));
    useY = normv2 >= tol;
    v3(useY,:) = bsxfun(@rdivide,bsxfun(@times,v3(useY,:),sign(v3(useY,2))),normv2(useY));
        
    if ~all(useY)
        w2 = e3(~useY,:) - bsxfun(@times,unitNormals(~useY,3),unitNormals(~useY,:)) - bsxfun(@times,v2(~useY,3),v2(~useY,:));
        normw2 = sqrt(sum(w2.^2,2));
        assert(all(normw2 >= tol),'rounding error? maybe tol is too strikt?')
        w2 = bsxfun(@rdivide,bsxfun(@times,w2,sign(w2)),normw2);
        v3(~useY,:) = w2;
    end

    unitNormalPerps = [unitNormals(cf,:);v2(cf,:);v3(cf,:)];
    unitNormalPerps = bsxfun(@times, unitNormalPerps, [sgn;sgn;sgn]);
    
    Nmatrix  = bsxfun(@times, unitNormalPerps(nF+1:end,:), [faceAreas(cf);faceAreas(cf)]);
    Nmatrix  = [bsxfun(@times,G.faces.normals(cf,:),sgn);Nmatrix];
    G.CMFD.N = Nmatrix;
    
    clear Nmatrix e1 e2 e3 v2 v3 w1 w2 w3 normv1 normv2 normw1 normw2 useX useY

% ------------------------------------------------------------------------
% Compute the curvature constant sigma for each face:

%     unitNormals = bsxfun(@rdivide, G.faces.normals , G.faces.areas); 
    unitSubNormals = bsxfun(@rdivide, subNormals, subAreas);
    
    sigma1 = bsxfun(@rdivide, sqrt(sum((unitNormals(faceNo,:) - unitSubNormals).^2,2)) , sqrt(faceAreas(faceNo)));

    sigma = max(bsxfun(@times, localEdge2Face, sigma1),[],1).';
    sigmaF = full(sigma(cf));

    G.CMFD.curvature = sigmaF;
    G.faces.curvature = full(sigma); % Only used in updateSigmastar.m
    
    if useCurvedBC
        isbc = any(G.faces.neighbors == 0,2);
        G.CMFD.cuF  = sigmaF > sigmastar + tol;
        G.faces.cufg = full(sigma) > sigmastar + tol;

        curvedBC = and(G.faces.cufg,isbc);
        G.faces.curvedBC = curvedBC(G.faces.cufg);
        clear curvedBC
    else
        isbc = any(G.faces.neighbors == 0,2);
        G.CMFD.cuF  = sigmaF > sigmastar + tol;
        if forceOnFlat
            G.CMFD.cuF  = true(size(sigmaF)); % Used to force extra degrees of freedom also on flat faces.
        end
        if any(G.CMFD.cuF(isbc(cf)))
            warning('does not use curved at bc')
            G.CMFD.cuF(isbc(cf)) = false;
        end
        G.faces.cufg = full(sigma) > sigmastar + tol;
        if forceOnFlat
            G.faces.cufg = true(size(sigma));
        end
        if any(G.faces.cufg(isbc))
            warning('does not use curved at bc')
            G.faces.cufg(isbc) = false;
        end
    end
    

% -------------------------------------------------------------------------    
% Compute C Matrix.

    AdotN = sum(rldecode(unitNormalPerps,repmat(diffN(cf),3,1)) .* ...
                    repmat(bsxfun(@times,subNormals(subcf,:), subsgn),3,1), 2);
    subCentVecs = subCentroids(subcf,:) - rldecode(G.cells.centroids(cellno,:),diffN(cf));

    C  = bsxfun(@times, repmat(subCentVecs,3,1), AdotN);
    C1 = C(1:end/3,:); C2 = C(end/3+1:2*end/3,:); C3 = C(2*end/3+1:end,:);
    C  = [sumUpOverSubFaces' * C1; sumUpOverSubFaces' * C2; sumUpOverSubFaces' * C3];
    C  = bsxfun(@rdivide, C, [faceAreas(cf);faceAreas(cf);faceAreas(cf)]);

    areaDivNormN = G.faces.areas(cf)./sqrt(sum(G.faces.normals(cf,:).^2,2));
%     normNDivArea = sqrt(sum(G.faces.normals(cf,:).^2,2))./G.faces.areas(cf);  Not in use
    
    C(1:nF,:) = bsxfun(@times, C(1:nF,:), areaDivNormN);

    G.CMFD.C = C;

% -------------------------------------------------------------------------    
    
    G.type = [G.type, { mfilename }];

end