function S = computeCurMimeticIP(G, rock, varargin)
% Compute curved mimetic inner product matrices.
%
% SYNOPSIS:
%   S = computeCurMimeticIP(G, rock)
%   S = computeCurMimeticIP(G, rock, 'pn', pv, ...)
%
% PARAMETERS:
%   G       - Grid structure as described by grid_structure.
%
%   rock    - Rock data structure with valid field 'perm'.  The
%             permeability is assumed to be in measured in units of
%             metres squared (m^2).  Use function 'darcy' to convert from
%             (milli)darcies to m^2, e.g.,
%
%                 perm = convertFrom(perm, milli*darcy)
%
%             if the permeability is provided in units of millidarcies.
%
%             The field rock.perm may have ONE column for a scalar
%             permeability in each cell, TWO/THREE columns for a diagonal
%             permeability in each cell (in 2/3 D) and THREE/SIX columns
%             for a symmetric full tensor permeability.  In the latter
%             case, each cell gets the permeability tensor
%
%                 K_i = [ k1  k2 ]      in two space dimensions
%                       [ k2  k3 ]
%
%                 K_i = [ k1  k2  k3 ]  in three space dimensions
%                       [ k2  k4  k5 ]
%                       [ k3  k5  k6 ]
%
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%               - Type -- The kind of system to assemble.  The choice made
%                         for this option influences which pressure solvers
%                         can be employed later.
%                         String.  Default value = 'hybrid'.
%                 Supported values are:
%                   - 'hybrid'      : Hybrid system with inverse of B
%                                     (for Schur complement reduction)
%
%               - InnerProduct -- The inner product with which to define
%                                 the mass matrix.
%                                 String.  Default value = 'ip_simple'.
%                 Supported values are:
%                   - 'ip_simple'   : inner product having the 2*tr(K)-term.
%                   - 'ip_tpf'      : inner product giving method equivalent
%                                     to two-point flux approximation (TPFA).
%                   - 'ip_quasitpf' : inner product ''close to'' TPFA
%                                     (equivalent for Cartesian grids with
%                                      diagonal permeability tensor).
%                   - 'ip_rt'       : Raviart-Thomas for Cartesian grids,
%                                     else not valid.
%                   - 'ip_quasirt'  : 
%
%                   - 'ip_dimfamily'  :  
%
%               - Inner_Param -- Extra parameter to inner product family
%                            dimfamily.
%
%
%               - Verbose -- Whether or not to emit progress reports during
%                            the assembly process.
%                            Logical.  Default value is dependent upon
%                            global MRST setting 'mrstVerbose'.
%
%               - facetrans  If facetrans is specified, the innerproduct is
%                            modified to account for a face
%                            transmissibilities specified as
%                                   [faces, face_transmissibilities]
%
% RETURNS:
%   S - Pressure linear system structure having the following fields:
%         - BI / B : inverse of B / B in hybrid/mixed system
%         - type   : system type (hybrid or mixed)
%         - ip     : inner product name
%
% COMMENTS:
%  In the hybrid discretization, the matrices B, C and D appear as
%       [  B   C  D  ]
%       [  C'  O  O  ]
%       [  D'  O  O  ]
%
% SEE ALSO:
%   solveIncompFlowCMFD, darcy, permTensor, mrstVerbose.

%{
Copyright 2012 Ruben Bø.

This file is part of the Curved Mimetic Finite-Difference module (CMFD).
CMFD is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

CMFD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CMFD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CMFD.  If not, see <http://www.gnu.org/licenses/>.

%}

% This file is largly based on the MRST file computeMimeticIP.m, with
% the following copyright
%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
% Based on:
% $Date: 2011-02-22 13:45:15 +0100 (Tue, 22 Feb 2011) $
% $Revision: 6992 $
%}



error(nargchk(2, inf, nargin, 'struct'));

opt = struct('InnerProduct', 'ip_simple', ...
             'verbose',      mrstVerbose, ...
             'Type',         'hybrid',    ...
             'facetrans',    zeros(0, 2),...
             'inner_param',   6);
opt = merge_options(opt, varargin{:});


dim = size(G.nodes.coords, 2);
%--------------------------------------------------------------------------
%- Model and discretisation parameters ------------------------------------
%
perm = permTensor(rock, dim);
assert (size(perm,1) == G.cells.num, ...
       ['Permeability must be defined in active cells only.\n', ...
        'Got %d tensors, expected %d (== number of cells).'],   ...
        size(perm,1), G.cells.num);


[systemType, ip, ipname, computeIP, computeInverseIP] = setup(opt);

%--------------------------------------------------------------------------
%- Compute discrete gradient (C) and flux continuity (D) matrices ---------
%
cf = double(G.cells.faces(:,1));

if opt.verbose,
   fprintf('Computing cell inner products ...\t\t')
   t0 = tic;
else
   t0 = [];
end


%--------------------------------------------------------------------------
%- Derive geometric primitives per half-contact in given model ------------
%
areas      = G.faces.areas    ([cf;cf;cf]);


%--------------------------------------------------------------------------
%- Preallocate SPARSE storage for 'BI' and/or 'B' -------------------------
%
dimProd = double(diff(G.cells.facePos));
dimProdb = zeros(G.cells.num,1);
sizVec  = sum(dimProd .^ 2);

if computeIP,        val_ip    = zeros([sizVec, 1]); end
% if computeInverseIP, val_invip = zeros([sizVec3, 1]); end
if computeInverseIP
    val_invip = cell(1,9);
    for i = 1:9
        val_invip{i} = zeros([sizVec, 1]); 
    end
end

% Current index to row/column in BI (or B), and current index to value.
ix = 0; ix2 = 0; ix2a = 0; ix2b = 0;

% Compute half-face transmissibilities from face transmissibilities.
if ~isempty(opt.facetrans),
   if ~all(opt.facetrans(:,2) > 0),
      dispif(opt.verbose, ...
             '\n\tAll face transmissibilities should be positive.\n\t\t\t');
   end
   error('This MRST functionality is not tested for CMFD')
   hft = zeros(G.faces.num, 1);
   hft(opt.facetrans(:,1)) = opt.facetrans(:,2);
   hft = hft.*sum(G.faces.neighbors ~= 0, 2);
   i=hft~=0; hft(i) = 1./hft(i);
   hft = hft(G.cells.faces(:,1));
else
   hft = zeros(size(G.cells.faces(:,1)));
end

%--------------------------------------------------------------------------
% Main loop -- Compute BI and/or B for each cell in model -----------------
%

Cmatrix     = G.CMFD.C;
Nmatrix     = G.CMFD.N;
cuF         = G.CMFD.cuF;
ncuF        = sum(cuF);

S.ncuF      = ncuF;
S.cuF       = cuF;


Nftot = double(G.cells.facePos(end)-1);
assert(size(Cmatrix,1)/3 == Nftot);
Ncells = G.cells.num;
for k = 1 : Ncells,
   nF  = dimProd(k);
   pR  = ix  + 1 : ix  + nF;   % half-face indices
   
   pRcu = pR(cuF(pR));
   nFcu = length(pRcu);
   dimProdb(k) = nFcu;
   utvalg = [pR, pRcu + Nftot, pRcu + 2*Nftot];
   pR2 = ix2 + 1 : ix2 + nF^2;
   pR2a = ix2a + 1 : ix2a + nF*nFcu;
   pR2b = ix2b + 1 : ix2b + nFcu^2;

   a = areas(utvalg); 
   v = G.cells.volumes(k);

   K = reshape(perm(k,:), [dim, dim]);
   C = Cmatrix(utvalg,:);
   N = Nmatrix(utvalg,:);

% This is a measure of the accuracy of the C'N = diag(volume) condtion   
%    if opt.verbose
%        if norm(C'*N - v*eye(3))/norm(eye(3)*v) > 10000*eps
%            fprintf(['\n rel err in C^T N = diag(v): ', num2str(norm(C'*N - v*eye(3))/norm(eye(3)*v)), ' # of cuf: ', num2str(nFcu)]);
%        end
%    end

   % Compute inner product if caller specified 'mixed' method.
   if computeIP,
      error('This MRST functionality is not tested for CMFD')
      val_ip(pR2)    = ip(a, v, K, C, N, false) + diag(hft(pR));
      
   end

   % Compute inverse inner product if caller specified 'hybrid' method.
   if computeInverseIP,
      this_ip = ip(a, v, K, C, N, true);
      
% This is a measure of the accuracy of the MNK = C condition      
%       if opt.verbose
%           if norm(N*K - this_ip*C)/norm(N*K) > 10000*eps
%               fprintf(['\n relative error NK - M^-1 C: ', num2str(norm(N*K - this_ip*C)/norm(N*K)), ' # of cuf: ', num2str(nFcu)]);
%           end
%       end

      if any(hft(pR)),         
         error('This MRST functionality is not tested for CMFD')
         val_invip(pR2) = inv( inv(this_ip) + diag(hft(pR)) );
      else
         val_invip{1}(pR2) = this_ip(1:nF,1:nF);
         val_invip{2}(pR2a) = this_ip(1:nF,nF+1:nF+nFcu);
         val_invip{3}(pR2a) = this_ip(1:nF,nF+nFcu+1:end);
         
         val_invip{4}(pR2a) = this_ip(nF+1:nF+nFcu,1:nF);
         val_invip{5}(pR2b) = this_ip(nF+1:nF+nFcu,nF+1:nF+nFcu);
         val_invip{6}(pR2b) = this_ip(nF+1:nF+nFcu,nF+nFcu+1:end);
         
         val_invip{7}(pR2a) = this_ip(nF+nFcu+1:end,1:nF);
         val_invip{8}(pR2b) = this_ip(nF+nFcu+1:end,nF+1:nF+nFcu);
         val_invip{9}(pR2b) = this_ip(nF+nFcu+1:end,nF+nFcu+1:end);
      end
   end

   ix  = ix  + nF;
   ix2 = ix2 + nF.^2;
   ix2a = ix2a + nF*nFcu;
   ix2b = ix2b + nFcu^2;
end
tocif(opt.verbose, t0);

%--------------------------------------------------------------------------
%- Assemble final block diagonal BI and/or B for complete model -----------
%

[ind1, ind2] = blockDiagIndex(dimProd, dimProd); 
[ind1a1, ind2a1] = blockDiagIndex(dimProd, dimProdb); 
[ind1a2, ind2a2] = blockDiagIndex(dimProdb, dimProd); 
[ind1b, ind2b] = blockDiagIndex(dimProdb, dimProdb); 

% ind1cu = ind1(cuFlog); ind2cu = ind2(cuFlog); % Denne stemmer nok ikke!!

clear areas dimProd centVecs outNormals perm
if opt.verbose,
   fprintf('Assembling global inner product matrix ...\t')
   t0 = tic;
else
   t0 = [];
end

if computeIP,
   n    = size(G.cells.faces, 1);
   S.B  = sparse(ind1, ind2, val_ip,    n, n);
end
if computeInverseIP,
   n    = size(G.cells.faces, 1); 
   BI = cell(1,9);
   
   BI{1} = sparse(ind1, ind2, val_invip{1}, n, n);
   BI{2} = sparse(ind1a1, ind2a1, val_invip{2}(1:ix2a), n, ncuF);
   BI{3} = sparse(ind1a1, ind2a1, val_invip{3}(1:ix2a), n, ncuF);

   BI{4} = sparse(ind1a2, ind2a2, val_invip{4}(1:ix2a), ncuF, n);
   BI{5} = sparse(ind1b, ind2b, val_invip{5}(1:ix2b), ncuF, ncuF);
   BI{6} = sparse(ind1b, ind2b, val_invip{6}(1:ix2b), ncuF, ncuF);

   BI{7} = sparse(ind1a2, ind2a2, val_invip{7}(1:ix2a), ncuF, n);
   BI{8} = sparse(ind1b, ind2b, val_invip{8}(1:ix2b), ncuF, ncuF);
   BI{9} = sparse(ind1b, ind2b, val_invip{9}(1:ix2b), ncuF, ncuF);
   
   S.BI =  [BI{1}, BI{2}, BI{3};
            BI{4}, BI{5}, BI{6};
            BI{7}, BI{8}, BI{9}];
end
tocif(opt.verbose, t0)

if opt.verbose && computeIP && computeInverseIP,
   fprintf('Max error in inverse = %g\n', ...
           norm(S.BI*S.B - speye(size(G.cells.faces,1)), inf));
end

%--------------------------------------------------------------------------
%- Define meta data on pressure system component matrices -----------------
%
S.ip    = ipname;
S.type  = systemType;


%--------------------------------------------------------------------------
% Individual inner product implementations
%--------------------------------------------------------------------------

function W = ip_simple(a, v, K, C, N, computeInverseIP)
% These are not mutually inverse for general cells

t = 6 * sum(diag(K)) / size(K,2);

if computeInverseIP,
   Q  = orth(bsxfun(@times, C, a));
   U  = eye(length(a)) - Q*Q';
   d  = diag(a);
   W  = (N*K*N' + t*(d * U * d)) ./ v;
else
   Q  = orth(bsxfun(@rdivide, N, a));
   U  = eye(length(a)) - Q*Q';
   di = diag(1 ./ a);
   W  = (C * (K \ C'))./v + (v / t)*(di * U * di);
end

%--------------------------------------------------------------------------

function W = ip_tpf(a, v, K, C, N, computeInverseIP) 
td = sum(C .* (N*K), 2) ./ sum(C.*C, 2);
if computeInverseIP,
   W = diag(abs(td));
else
   W = diag(1 ./ abs(td));
end

%--------------------------------------------------------------------------

function W = ip_rt(a, v, K, C, N, computeInverseIP)
% Only for orthogonal normals.
Ki = inv(K);
u  = diag(diag(Ki)) .* (v / 6);
di = diag(1 ./ a);
Q  = abs(N')*di/sqrt(2);
n = abs(N*N');
mask = n>eps*max(n(:));
W  = mask.*(C*(K\C')./v + di*Q'*u*Q*di);

if computeInverseIP, W = inv(W); end

%--------------------------------------------------------------------------

function W = ip_quasirt(a, v, K, C, N, computeInverseIP) 
% Applicable to general grids.  Coincides with ip_rt for orthogonal grids.

t = 6;

W = N * K * N';

if computeInverseIP,
   Q  = orth(C);
   U  = eye(length(a)) - Q*Q';
   d  = diag(diag(W));
   W  = (W + t*(U * d * U)) ./ v;
else
   Q  = orth(N);
   U  = eye(length(a)) - Q*Q';
   di = diag(1 ./ diag(W));
   W  = (C * (K \ C'))./v + (v / t)*(U * di * U);
end

%--------------------------------------------------------------------------

function W = ip_dimfamily(a, v, K, C, N, computeInverseIP, t)
% Applicable to general grids.  Coincides with ip_rt for orthogonal grids.

W = N * K * N';

if computeInverseIP,
   Q  = orth(C);
   U  = eye(length(a)) - Q*Q';
   d  = diag(diag(W));
   W  = (W + t*(U * d * U)) ./ v;
else
   Q  = orth(N);
   U  = eye(length(a)) - Q*Q';
   di = diag(1 ./ diag(W));
   W  = (C * (K \ C'))./v + (v / t)*(U * di * U);
end

%--------------------------------------------------------------------------

function W = ip_quasitpf(a, v, K, C, N, computeInverseIP)
Q  = orth(C);
U  = eye(length(a)) - Q*Q';
W1 = N * K * N';

W  = (W1 + 2*U*diag(diag(W1))*U) ./ v;

if ~computeInverseIP, W = inv(W); end

%--------------------------------------------------------------------------
% Other private utility functions below.
%--------------------------------------------------------------------------

function [Type, ipfun, ipName, IP, invIP] = setup(opt)

ipName  = opt.InnerProduct;
Type    = opt.Type;

switch lower(Type),
   case 'hybrid',         [invIP, IP] = deal(true , false);
   case {'mixed','tpfa'}, [invIP, IP] = deal(false, true );
   case 'comp_hybrid',    [invIP, IP] = deal(true , true );
   otherwise,
      error(id('SystemType:Unknown'), ...
            'Unkown system type ''%s''.', Type);
end

ipNames = {'ip_simple', 'ip_tpf', 'ip_quasitpf', 'ip_rt', ...
           'ip_quasirt', 'ip_dimfamily'};

if ~any(strcmp(ipName, ipNames)),
   error(id('InnerProduct:Unknown'), ...
         'Unknown inner product ''%s''.', ipName);
end

if opt.verbose && strcmp(ipName, 'ip_rt'),
   disp(['NOTE: The mixed MFEM inner product is only valid for', ...
         ' Cartesian grids as given by cartGrid.'])
end

dispif(opt.verbose, 'Using inner product: ''%s''.\n', ipName);
switch lower(ipName),
   case 'ip_simple'   , ipfun = @ip_simple;
   case 'ip_tpf'      , ipfun = @ip_tpf;
   case 'ip_quasitpf' , ipfun = @ip_quasitpf;
   case 'ip_rt'       , ipfun = @ip_rt;
   case 'ip_quasirt'  , ipfun = @ip_quasirt;
   case 'ip_dimfamily', ipfun = ...
                        @(varargin) ip_dimfamily(varargin{:}, ...
                                                 opt.inner_param);
   otherwise,
      % Can't happen.  'ipName' is already verified.
      error(id('InnerProduct:Unknown'), ...
            'Unknown inner product ''%s''.', ipName);
end

%--------------------------------------------------------------------------

function s = id(s)
s = ['computeMimeticIP:', s];
