function G = perturbate(G,Lx,Ly,Lz,ComputeGeometry,alpha)
% Perturbe the nodes of a cartesian grid witout changing the global geometry
%
% SYNOPSIS:
%   G = perturbate(G, Lx, Ly, Lz, ComputeGeometry, alpha)
%   G = perturbate(G, Lx, Ly, Lz)
%   G = perturbate(G)
% 
% PARAMETERS:
%   G       - Grid structure as described by grid_structure.
%
%   Lx      - size of grid in x-direction in meters
%   Ly      - size of grid in y-direction in meters
%   Lz      - size of grid in z-direction in meters
%
%   ComputeGeometry - logical. Weather or not to run computeGeometry.m
%
%   alpha   - numerical value for how much the grid should be
%             perturbed
%
% RETURNS
%   G       - Grid structure with same topology and global faces, but with
%             perturbed nodes

%{
Copyright 2012 Ruben Bø.

This file is part of the Curved Mimetic Finite-Difference module (CMFD).
CMFD is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

CMFD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CMFD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CMFD.  If not, see <http://www.gnu.org/licenses/>.

%}

    if nargin == 1
        Lx = G.cartDims(1); Ly = G.cartDims(2); Lz = G.cartDims(3);
        ComputeGeometry = 1;
        alpha = 0.5;
    elseif nargin == 4
        ComputeGeometry = 1;
        alpha = 0.5;
    end
    
    % Perturbating the inner nodes
    c = G.nodes.coords;
    J = or( or( any(c==0, 2),       any(c(:,1)==Lx,2) )  , ...
            or( any(c(:,2)==Ly,2) , any(c(:,3)==Lz,2))     );  
        
%     rng(1);
    rand('state',1)
    G.nodes.coords(~J,:) = c(~J,:) + alpha.*(0.6*rand(sum(~J),3) - 0.3);
    
    % Perturbating the sides
    is =  [1,  1,  1,  1,  2,  2  ];
    Nis = [Lx, Lx, Lx, Lx, Ly, Ly ];
    js =  [2,  2,  3,  3,  3,  3  ];
    Njs = [Ly, Ly, Lz, Lz, Lz, Lz ];
    ks =  [3,  3,  2,  2,  1,  1  ];
    Nks = [0,  Lz, 0,  Ly, 0,  Lx ];
    
    for t=1:6   
        J1 = or( or( and(any(c(:,is(t))==0,2) , c(:,ks(t))==Nks(t)) , and(any(c(:,js(t))==0,2) , c(:,ks(t))==Nks(t)) ) , ... 
                 or( and(any(c(:,is(t))==Nis(t),2) , c(:,ks(t))==Nks(t)) , and(any(c(:,js(t))==Njs(t),2) , c(:,ks(t))==Nks(t)) ));
        J2 = and( any(c(:,ks(t))==Nks(t),2) , ~J1 ); % This is all nodes on the interior of one face
        
        G.nodes.coords(J2,[is(t) js(t)]) = c(J2,[is(t) js(t)]) + alpha.*(0.6*rand(sum(J2),2) - 0.3);
    end
    
    G.type = [G.type, { mfilename }];
    
    if ComputeGeometry
        G = computeGeometry(G);
    end
    

end

