%% Tutorial for the CMFD module
%
% This tutorial is largly based on the MRST tutorial found here:
% http://www.sintef.no/Projectweb/MRST/Tutorials/Flow-Solver-Tutorial/

%%
% The purpose of this example is to give an overview of how to set up and
% use the single-phase curved  mimetic pressure solver to solve the
% single-phase pressure equation
%
% $$\nabla \cdot v = q, \quad v = \textbf{--}K/\mu \nabla p$$
%
% for a flow driven by Dirichlet and Neumann boundary conditions. Our
% geological model will be simple a Cartesian grid with isotropic,
% homogeneous permeability.

%% Define geometry
%
% Construct a Cartesian grid of size 10-by-10-by-4 cells, where each cell
% has dimension 1-by-1-by-1. Because our flow solvers are applicable for
% general unstructured grids, the Cartesian grid is here represented using
% an unstructured format, in which cells, faces, nodes, etc. are given
% explicitly. Then the nodes of the grid is perturbed so that the inner
% faces are curved, but the global shape of the reservoir is unchanged.

nx = 10; ny = 10; nz = 4;

G = cartGrid([nx, ny, nz]);
G = perturbate(G,nx,ny,nz,false,0.5);
display(G);

%% Plot geometry

plotGrid(G);
view(3), camproj orthographic, axis tight, camlight headlight

%% Process geometry
%
% Having set up the basic structure, we continue to compute centroids and
% volumes of the cells and centroids, normals, and areas for the faces.
% For a Cartesian grid, this information can trivially be computed, but is
% given explicitly so that the flow solver is compatible with fully
% unstructured grids.

G = computeGeometry(G);

%%
% The mimetic method uses one degree of freedom for the flux over each
% face. In the curved mimetic method, two extra degrees of freedom for the
% velocity are added for each face. These new unknowns are associated with
% the flow orthogonal to the normal direction. We therefor compute
% orthogonal vectors 'a2' and 'a3' for each face. A plot of a single grid
% cell and the normal and orthogonal vectors are shown and the relevant
% constants needed are calculated.

tutorialGridCell

G = computeCmatrix(G);

%% Define rock and fluid properties
%
% The only parameters in the single-phase pressure equation are the
% permeability  and the fluid viscosity. We set the permeability to be
% homogeneous and isotropic.
%
% The viscosity is specified by saying that the reservoir is filled with
% a single fluid, for which the default viscosity value equals unity. Our
% flow solver is written for a general incompressible flow and requires
% the evaluation of a total mobility, which is provided by the fluid
% object.

rock.perm = repmat(100*milli*darcy(), [G.cells.num, 1]);

fluid     = initSingleFluid('mu' ,    1*centi*poise     , ...
                            'rho', 1014*kilogram/meter^3);

%%
% For heterogeneous permeability we calculate the constants needed for the
% new continuity condition. This step is not necessary for the homogeneous
% case, where the naive continuity condition for the new unknowns is valid
% (and the two continuity conditions are the same).
% Note that this curved mimetic method is only valid for permeabilites on
% the form $K_i = \alpha K_j$, which includes the isotropic case and cases
% of anisotropy whith the same anisotropy rate throughout the reservoir.

G = makeKalpha(G, rock);

%% Initialize reservoir simulator
%
% To simplify communication among different flow and transport solvers,
% all unknowns are collected in a structure. Here this structure is
% initialized with uniform initial reservoir pressure equal 0 and
% (single-phase) saturation equal 0.0 (using the default behavior of
% initResSol)

resSol = initResSol(G, 0.0);
display(resSol);

%% Impose Dirichlet boundary conditions
%
% Our flow solvers automatically assume no-flow conditions on all outer
% (and inner) boundaries; other type of boundary conditions need to be
% specified explicitly.
%
% Here, we impose Neumann conditions (flux of 1 m^3/day) on the global
% left-hand side. The fluxes must be given in units of m^3/s, and thus we
% need to divide by the number of seconds in a day. Similarly, we set
% Dirichlet boundary conditions p = 0 on the global right-hand side of the
% grid, respectively. For a single-phase flow, we need not specify the
% saturation at inflow boundaries. Similarly, fluid composition over
% outflow faces (here, right) is ignored by pside.

bc = fluxside([], G, 'LEFT',  1*meter^3/day());
bc = pside   (bc, G, 'RIGHT', 0);
display(bc);

%% Construct linear system
%
% Construct mimetic pressure linear system components for the system $Ax =
% b$ in which $B_1$, $C$, and $D_1$ are the same as for the mimetic method
% and $E_1 = D_1$. For the homogeneous case, $E_2 = D_2$ and $E_3 = D_3$,
% but for the new continuity condition, the matrices $E_2$ and $E_3$
% contain the ratio of permeability, computed by <matlab:doc('makeKalpha')
% |makeKalpha|>. Furthermore $v_1$ is the flux and $\mathit{cp}_1$ the face
% pressures, as in the traditional mimetic method. The quantities $v_2$,
% $v_3$, $\mathit{cp}_2$ and $\mathit{cp}_3$ are new unknowns associated
% with the orthogonal vectors $a_2$ and $a_3$ as in the figure from
% <matlab:edit('tutorialGridCell') |tutorialGridCell|>.
%
% $$ A = \left[\begin{array}{ccccccc}
% B_1 &  0  &  0  & C & D_1 &  0  & 0   \\
%  0  & B_2 &  0  & 0 &  0  & D_2 & 0   \\
%  0  &  0  & B_3 & 0 &  0  &  0  & D_3 \\
%  C' &  0  &  0  & 0 &  0  &  0  & 0   \\
% E_1'&  0  &  0  & 0 &  0  &  0  & 0   \\
%  0  & E_2'&  0  & 0 &  0  &  0  & 0   \\
%  0  &  0  & E_3'& 0 &  0  &  0  & 0\end{array}\right]
% \left[\begin{array}{c}
% v_1 \\ v_2 \\ v_3 \\ -p \\
% \mathit{cp}_1 \\
% \mathit{cp}_2 \\
% \mathit{cp}_3\end{array}\right]
% = \left[\begin{array}{c}
% f_1 \\ f_2 \\ f_3 \\ g \\ h_1 \\ h_2 \\ h_3\end{array}\right] = b.$$
%
% The coefficient matrix $A$ and the system right-hand side vector $b$ are
% constructed based on input grid and rock properties for the case with no
% gravity.

gravity off;
S = computeCurMimeticIP(G, rock);

%% Plot the structure of the coefficient matrix
% (here we use BI, the inverse of B, rather than B)
clf, subplot(1,2,1)
cellNo  = rldecode(1:G.cells.num, diff(G.cells.facePos), 2) .';
C       = sparse(1:numel(cellNo), cellNo, 1);
D       = sparse(1:numel(cellNo), double(G.cells.faces(:,1)), 1, ...
                 numel(cellNo), G.faces.num);

C       = [C;zeros(2*S.ncuF,G.cells.num)];
D       = blkdiag(D,D(S.cuF,G.faces.cufg),D(S.cuF,G.faces.cufg));

spy([S.BI               , C        , D        ; ...
     C', zeros(size(C,2), size(C,2) + size(D,2)); ...
     D', zeros(size(D,2), size(C,2) + size(D,2))]);
title('Hybrid pressure system matrix')

%%
% The block structure can clearly be seen in the sparse matrix |A|, which
% is never formed in full. Indeed, rather than storing |B|, we store its
% inverse $B^{\textbf{--}1}$. Similarly, the |C| and |D| blocks are not
% represented in the |S| structure; they can easily be formed explicitly
% whenever needed, or their action can easily be computed.

display(S);

%% Solve the linear system
%
% Solve linear system construced from |S| and |bc| to obtain solution for
% flow and pressure in the reservoir. The option |MatrixOutput=true| adds
% the system matrix |A| to |resSol| to enable inspection of the matrix.

resSol = solveIncompFlowCMFD(resSol, G, S, fluid, ...
                         'bc', bc, 'MatrixOutput', true);
display(resSol);

%% Inspect results
%
% The |resSol| object contains the Schur complement matrix used to solve
% the hybrid system.

subplot(1,2,2), spy(resSol.A);
title('Schur complement system matrix');

%% Plot results
% We then plot convert the computed pressure to unit 'bar' before plotting
% result.

figure
plotCellData(G, convertTo(resSol.pressure(1:G.cells.num), barsa()), ...
             'EdgeColor', 'k');
title('Cell Pressure [bar]')
xlabel('x'), ylabel('y'), zlabel('Depth');
view(3); shading faceted; camproj perspective; axis tight;
colorbar
