function G = wellTest
% In this example, there is a choice of three different grid versions, a
% regular flat grid, a curved grid, and a grid with perturbed inner nodes.
% One can also choose to run the mimetic, the curved mimetic or both
% methods.


    %% Definitions for size of grid and placement of sources
    nx = 19; ny = 19; nz = 5;

    dx = 1; dy = 1;
    Lx = dx*nx; Ly = dy*ny; Lz = nz;
    din = 3; doutx = 4; douty = din; % number of cells from mid cell. Must be >= 1.
    assert(mod(nx,2) == 1, 'number of cells in x-direction must be odd.')
    assert(mod(ny,2) == 1, 'number of cells in y-direction must be odd.')
    assert(mod(nz,2) == 1, 'number of cells in z-direction must be odd.')

    
    % Choose what version of the example you want to run:
    useM = true;
    useC = true;
    grids = {'flat', 'curved', 'perturbed'};
    grid = grids{1}; % Change this to run the different grids
    
    %% Make the model grid
    forceOnFlat = false;
    XX = 0:Lx/nx:Lx;
    YY = 0:Ly/ny:Ly;
    switch grid
        case 'flat'
            grdecl = cartesianGrdecl(dx*XX, dy*YY, 0:Lz/nz:Lz);
            G = processGRDECL(grdecl);
            forceOnFlat = true;
        case 'curved'
            alpha = 2*pi/3; beta = 3*pi/5;
            [X, Y] = ndgrid(linspace(0, alpha, nx+1), linspace(0, beta, ny+1));
            Z = 0.5*sin(X).*sin(Y);
            grdecl = cartesianGrdecl(0:Lx/nx:Lx, 0:Ly/ny:Ly, 0:Lz/nz:Lz, 'depthz', Z);
            
            [~,~,z] = buildCornerPtNodes(grdecl);
            z(:,:,1) = 0;
            cutoff = grdecl.cartDims(3);
            Idx = z > cutoff;
            z(Idx) = cutoff;
            z = sort(z,3,'ascend');
            grdecl.ZCORN = z(:);
            
            G = processGRDECL(grdecl);
        case 'perturbed'
            grdecl = cartesianGrdecl(dx*XX, dy*YY, 0:Lz/nz:Lz);
            G = processGRDECL(grdecl);
            G = computeGeometry(G);
            G = perturbate(G,Lx,Ly,Lz,1,1);
    end
    G = computeGeometry(G);

    %% Define permeability and fluid
    K = 1000;
    rock.perm = repmat(K*milli*darcy, [G.cells.num, 1]);
%     rock.perm(1:nx*ny) = repmat(1*milli*darcy, [nx*ny, 1]);
%     rock.perm(nx*ny*(nz-1)+1:nx*ny*nz) = repmat(1*milli*darcy, [nx*ny, 1]);

    MU = 1;
    fluid = initSingleFluid('mu', MU*centi*poise, 'rho', 1014*kilogram/meter^3);
    
    
    %% Find indices for the source placements
    
    % Make a reference grid for logical numbering only:
    grdeclCART = cartesianGrdecl(XX, YY, 0:Lz/nz:Lz);   
    Gcart = processGRDECL(grdeclCART);    
    Gcart = computeGeometry(Gcart);
    
    
    xmid = Lx/2; ymid = Ly/2;
    At = ymid + dy*(din + 0.5); Ab = ymid + dy*(din - 0.5);
    Al = xmid - dx/2; Ar = xmid + dx/2;
    
    IA = and( ...
                and(Gcart.cells.centroids(:,1) > Al, Gcart.cells.centroids(:,1) < Ar), ...
                and(Gcart.cells.centroids(:,2) > Ab, Gcart.cells.centroids(:,2) < At));
    
    Bt = ymid - dy*(douty - 0.5); Bb = ymid - dy*(douty + 0.5);
    Bl = xmid - dx*(doutx + 0.5); Br = xmid - dx*(doutx - 0.5);
    Cl = xmid + dx*(doutx - 0.5); Cr = xmid + dx*(doutx + 0.5);
    
    IB = and( ...
                and(Gcart.cells.centroids(:,1) > Bl, Gcart.cells.centroids(:,1) < Br), ...
                and(Gcart.cells.centroids(:,2) > Bb, Gcart.cells.centroids(:,2) < Bt));
    
    IC = and( ...
                and(Gcart.cells.centroids(:,1) > Cl, Gcart.cells.centroids(:,1) < Cr), ...
                and(Gcart.cells.centroids(:,2) > Bb, Gcart.cells.centroids(:,2) < Bt));
            
    cells1 = 1:length(IA); cells1 = cells1(IA);
    cells2 = 1:length(IB); 
    cellB = cells2(IB); cellB = cellB(1);
    cellC = cells2(IC); cellC = cellC(1);
    cells2 = cells2(or(IB,IC));
    WELLS = or(IA,or(IB,IC));
    

    clear Gcart At Ab Al Ar Bt Bl Ct Cl IB IC IA grdecl grdeclCART
    %%
    
    plotGrid(G, 'facea', 0.3, 'edgea',0.1);    
    view(120,30), axis tight
    hold on, plotGrid(G, [cells1,cells2])
    
    
    %% Value and definition of the source terms
    val = 1;
    val1 = 2*val*ones(size(cells1));
    val2 = -1*val*ones(size(cells2));
    
    src = addSource([], cells1, val1);
    src = addSource(src, cells2, val2);


    %% Compute approximate analytical solution
    bc = pside([], G, 'XMin', 0); bc = pside(bc, G, 'XMax', 0); bc = pside(bc, G, 'YMin', 0); bc = pside(bc, G, 'YMax', 0);
    xcoords = G.cells.centroids([cells1(1),cellB,cellC],1);
    ycoords = G.cells.centroids([cells1(1),cellB,cellC],2);
    qs = [2*val,-val,-val];
    const = K*milli*darcy/(MU*centi*poise/1.014);
    [Pexfull, pexfaces] = exactPressure(G,xcoords,ycoords, -qs, const, bc.face);
      
    
    %% Set boundary conditions from analytical pressure solution
    BC = addBC([],bc.face,'pressure',pexfaces(bc.face));
    
    %% find mid section x-y plane
    Imid = and(G.cells.centroids(:,3) > floor(nz/2), G.cells.centroids(:,3) < ceil(nz/2));
    cmid = 1:length(Imid); cmidNoW = cmid(and(~WELLS,Imid)); cmid = cmid(Imid);
    
    %% Solvers
    if useM
        % Mimetic
        Smim = computeMimeticIP(G,rock);    
        state = initResSol(G,0);
        state = solveIncompFlow(state,G,Smim,fluid,'src', src, 'bc', BC);
        presM = state.pressure(and(~WELLS,Imid));

        clear state Smim
        

        % Plot of mimetic
        Pexmid = Pexfull(and(~WELLS,Imid));
        figure, subplot(1,2,1), plotCellData(G, (abs(presM-Pexmid))/(norm(Pexmid,inf)), cmidNoW, 'facea', 0.3, 'edgea',0.1)
        view(0,90), colorbar
        c=colormap('hot'); 
        c=flipud(c); 
        colormap(c) 
        caxis([0 2.0e-2])
        
    end
    
    if useC
        % Curved mimetic
        G = computeCmatrix(G,0,forceOnFlat,0);
        S = computeCurMimeticIP(G, rock, 'InnerProduct', 'ip_simple');    
        rSol = initResSol(G,0); 
        rSol = solveIncompFlowCMFD(rSol, G, S, fluid, 'src', src, 'bc', BC);
        pres = rSol.pressure(and(~WELLS,Imid));

        clear rSol S

        % Plot of curved mimetic
        Pexmid = Pexfull(and(~WELLS,Imid));
        subplot(1,2,2), plotCellData(G, (abs(pres-Pexmid))/(norm(Pexmid,inf)), cmidNoW, 'facea', 0.3, 'edgea',0.1)
        view(0,90), colorbar
        c=colormap('hot'); 
        c=flipud(c); 
        colormap(c) 
        caxis([0 2.0e-2])
        
    end
    

%     figure, plotCellData(G, (state.pressure(cmid)-rSol.pressure(cmid))/norm(rSol.pressure(cmid),inf), cmid, 'facea', 0.3, 'edgea',0.1)
%     view(0,0), colorbar

    
    %% Scatter plot:
    figure, subplot(1,2,1), scatter(Pexmid,presM)
    
    subplot(1,2,2), scatter(Pexmid,pres)
    
    %% Print L2-norm on screen
    if useC
        fprintf(['L2 norm of curved mimetic method vs approx analytical sol: ', ...
            num2str(norm(pres-Pexmid)/norm(Pexmid)),'\n']);
    end
    if useM
        fprintf(['L2 norm of mimetic method vs approx analytical sol:        ', ...
            num2str(norm(presM-Pexmid)/norm(Pexmid)),'\n']);
    end
    if useM && useC
        fprintf(['L2 norm of mimetic method vs the curved mimetic method:    ', ...
            num2str(norm(presM-pres)/norm(presM)),'\n']);
    end
end

function [p, pf] = exactPressure(G, xcoords, ycoords, qs, const, face)

    X = G.cells.centroids(:,1); Y = G.cells.centroids(:,2);
    p = zeros(G.cells.num,1);
    for i = 1:length(xcoords)
        p = p + qs(i)/(2*pi*const)*log(sqrt((X-xcoords(i)).^2 + (Y-ycoords(i)).^2));
    end
    
    Xf = G.faces.centroids(:,1); Yf = G.faces.centroids(:,2);
    pf = zeros(G.faces.num,1);
    for i = 1:length(xcoords)
        pf = pf + qs(i)/(2*pi*const)*log(sqrt((Xf-xcoords(i)).^2 + (Yf-ycoords(i)).^2));
    end 
    
    p  = p  - 2*min(pf(face));
    pf = pf - 2*min(pf(face));
        
end
