function G = makeKalpha(G, rock)
% Adding the permeability ratios 'kalpha' to the grid structure
%
% SYNOPSIS:
%   G = makeKalpha(G, rock)
%
% PARAMETERS:
%   G       - Grid structure as described by grid_structure.
%
%   rock    - Rock data structure with valid field 'perm'.  The
%             permeability is assumed to be in measured in units of
%             metres squared (m^2).  Use function 'darcy' to convert from
%             (milli)darcies to m^2, e.g.,
%
%                 perm = convertFrom(perm, milli*darcy)
%
%             if the permeability is provided in units of millidarcies.
%
%             In MRST, the field rock.perm may have ONE column for a scalar
%             permeability in each cell, THREE columns for a diagonal
%             permeability in each cell (in 3D) and SIX columns
%             for a symmetric full tensor permeability. But here, only
%             isotropic permeability is  implemented. 
%
% RETURNS:
%   G - Grid structure with added fields:
%         - CMFD
%             - kalpha   -- A Nf-by-1 array of the ratio of the
%                           permeabilities in the two neighbouring cells,
%                           where Nf = size(G.cells.faces,1).
%

%{
Copyright 2012 Ruben Bø.

This file is part of the Curved Mimetic Finite-Difference module (CMFD).
CMFD is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

CMFD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CMFD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CMFD.  If not, see <http://www.gnu.org/licenses/>.

%}

    %Check that the permeability satisfies K_i = \alpha K_j
    if size(rock.perm,2) ~= 1
        ratio = max(rock.perm,[],1);
        scaled = bsxfun(@ldivide,rock.perm,ratio);
        assert(all(all(scaled == repmat(scaled(:,1),1,size(scaled,2)))), ...
            'permeability must be isotropic');
    end

    faces = rldecode(1:G.cells.num, diff(G.cells.facePos), 2) .' ;
    kalpha = rock.perm(faces);
    kalpha = kalpha./norm(kalpha,inf);
    G.CMFD.kalpha = kalpha;

end