function [flux, press, lam, varargout] = schurComplementSymmCMFD(BI, C, ...
                                                        D, E, f, g, h, varargin)
%Solve symmetric system of linear eqns using Schur Complement analysis.
%
% SYNOPSIS:
%   [flux, press, lam]    = schurComplementSymm(BI, C, D, E, f, g, h)
%   [flux, press, lam]    = schurComplementSymm(BI, C, D, E, f, g, h, ...
%                                               'pn', pv, ...)
%   [flux, press, lam, S] = schurComplementSymm(...)
%
% DESCRIPTION:
%   Solves the symmetric, hybrid (block) system of linear equations
%
%       [  B   C   D  ] [  v ]     [ f ]
%       [  C'  0   0  ] [ -p ]  =  [ g ]
%       [  E'  0   0  ] [ cp ]     [ h ]
%
%   with respect to the half-face (half-contact) flux 'v', the cell
%   pressure 'p', and the face (contact) pressure 'cp'.  The system is
%   solved using a Schur complement reduction to a symmetric system of
%   linear equations
%
%        S cp = R          (*)
%
%   from which the contact pressure 'cp' is recovered.  Then, using back
%   substitution, the cell pressure and half-contact fluxes are
%   recovered.
%
%   Note, however, that the function assumes that any faces having
%   prescribed pressure values (such as Dirichlet boundary condtions or
%   wells constrained by BHP targets) have already been eliminated from
%   the blocks 'D', 'E' and 'h'.
%
% PARAMETERS:
%   BI      - Inverse of the matrix B in the block system.  Assumed SPD.
%   C       - Block 'C' of the block system.
%   D       - Block 'D' of the block system.
%   E       - Block 'E' of the block system.
%   f       - Block 'f' of the block system right hand side.
%   g       - Block 'g' of the block system right hand side.
%   h       - Block 'h' of the block system right hand side.
%   'pn'/pv - List of 'key'/value pairs defining optional parameters.  The
%             supported options are:
%               Regularize -- Whether or not to enforce p(1)==0 (i.e., set
%                             pressure zero level) which may be useful when
%                             solving pure Neumann problems.
%                             Logicial.  Default value = FALSE.
%               LinSolve   -- Function handle to a solver for the system
%                             (*) above.  Assumed to support the syntax
%
%                                    x = LinSolve(A, b)
%
%                             to solve a system Ax=b of linear equations.
%                             Default value: LinSolve = @mldivide
%                             (backslash).
%
% RETURNS:
%   flux  - The half-contact fluxes 'v'.
%   press - The cell pressure 'p'.
%   lam   - The contact pressure 'cp'.
%   S     - Schur complement reduced system matrix used in recovering
%           contact pressures.
%           OPTIONAL.  Only returned if specifically requested.
%
% SEE ALSO:
%   tpfSymm, mixedSymm, solveIncompFlow, solveIncompFlowMS.

%{
Copyright 2012 Ruben Bø.

This file is part of the Curved Mimetic Finite-Difference module (CMFD).
CMFD is a module for the MATLAB Reservoir Simulation Toolbox (MRST).

CMFD is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

CMFD is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with CMFD.  If not, see <http://www.gnu.org/licenses/>.

%}

% This file is largly based on the MRST file schurComplementSymm.m, with
% the following copyright
%{
Copyright 2009, 2010, 2011 SINTEF ICT, Applied Mathematics.

This file is part of The MATLAB Reservoir Simulation Toolbox (MRST).

MRST is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

MRST is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with MRST.  If not, see <http://www.gnu.org/licenses/>.
% $Date: 2011-02-22 13:45:15 +0100 (Tue, 22 Feb 2011) $
% $Revision: 6992 $
%}



opt = struct('Regularize', false, 'LinSolve', @mldivide);
opt = merge_options(opt, varargin{:});

regularize = opt.Regularize;

%--------------------------------------------------------------------------
% Schur complement analysis to reduce to SPD system for contact pressure --
%
ncell = numel(g);               % Number of cells in model

BIDf   = BI * [D, f];
CtBIDf = C' * BIDf;


BIEf2   = BI * [E, f];
EtBIDf2 = E' * BIDf;
CtBIEf2 = C' * BIEf2;


M  = CtBIDf(:, 1:end-1);         % == C' * inv(B) * D, == F1
M2 = CtBIEf2(:, 1:end-1);        % == C' * inv(B) * E, == F2
S  = EtBIDf2(:, 1:end-1);        % == E' * inv(B) * D

g = g - CtBIDf(:, end);          % == g - C'*inv(B)*f
h = h - EtBIDf2(:, end);         % == h - E'*inv(B)*f

L = diag(C.' * BI * C);
if regularize, L(1) = 2 * L(1); end

L      = sparse(1 : ncell, 1 : ncell, L);

LIMg   = L  \ [M, g];           % use MLDIVIDE for improved accuracy
M2tLIMg = M2' * LIMg;

S = S - M2tLIMg(:, 1:end-1);     % == S - M2.'*inv(L)*M
h = M2tLIMg(:,end) - h;          % == M2.'*inv(L)*g - h


%--------------------------------------------------------------------------
% Solve Schur system to recover unknown contact pressures -----------------
%   (E.'*BI*D - M2.'*inv(L)*M) cp = right hand side
%

lam   = opt.LinSolve(S, h);

%--------------------------------------------------------------------------
% Recover cell pressure from reduced (diagonal) system --------------------
%   L p = g + M*lam
%
press = LIMg(:, end) + LIMg(:, 1:end-1)*lam;

if issparse(press),
   % This happens whenever NUMEL(lam)==1
   press = full(press);
end

%--------------------------------------------------------------------------
% Recover half-contact fluxes from reduced system -------------------------
%   B v = f + C*p - D*lam
%
flux  = BIDf(:, end) - BIDf(:, 1:end-1)*lam + BI*C*press;

if nargout > 3, varargout{1} = S; end
